==> Smdn.Formats.Mime-0.32.tar.bz2: Smdn.Formats.Mime-0.32/Smdn.Formats.Mime/Smdn.Formats.Mime/MimeHeader.cs <==
// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2012 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;

using Smdn.Formats.Mime.Formatting;

namespace Smdn.Formats.Mime {
  public class MimeHeader : IHeaderFieldBody {
    internal const char NameBodyDelimiter = ':';

    public string Name {
      get { return name; }
    }

    public string Value {
      get; internal set;
    }

    public MimeEncodingMethod Encoding {
      get { return encoding; }
    }

    public Encoding Charset {
      get { return charset; }
    }

    internal MimeHeader(string name, string @value)
      : this(name, @value, MimeEncodingMethod.None, null)
    {
    }

    internal MimeHeader(string name, string @value, MimeEncodingMethod encoding, Encoding charset)
    {
      this.name = name;
      this.Value = @value;
      this.encoding = encoding;
      this.charset = charset;
    }

    public bool IsNameEquals(string name)
    {
      return string.Equals(this.name, name, StringComparison.OrdinalIgnoreCase);
    }

    public string GetParameter(string attr)
    {
      if (attr == null)
        throw new ArgumentNullException("attr");
      if (attr.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString("attr");

      string val;

      if (EnsureValueParsed(null).TryGetValue(attr, out val))
        return val;
      else
        return null;
    }

    public KeyValuePair<string, string>[] GetAllParameters()
    {
      var ret = new List<KeyValuePair<string, string>>();

      foreach (var pair in EnsureValueParsed(null)) {
        if (pair.Key.Length == 0)
          continue; // if header field

        ret.Add(pair);
      }

      return ret.ToArray();
    }

    public void ForEachParameter(Action<KeyValuePair<string, string>> action)
    {
      if (action == null)
        throw new ArgumentNullException("action");

      foreach (var pair in EnsureValueParsed(null)) {
        if (pair.Key.Length == 0)
          continue; // if header field

        action(pair);
      }
    }

    public bool HasParameter(string attr)
    {
      if (attr == null)
        throw new ArgumentNullException("attr");
      if (attr.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString("attr");

      return EnsureValueParsed(null).ContainsKey(attr);
    }

    public string GetValueWithoutParameter()
    {
      string val;

      if (EnsureValueParsed(null).TryGetValue(string.Empty, out val))
        return val;
      else
        return null;
    }

    /*
     * attr=""      quoted
     * attr*=       extended
     * attr*0=      sectioned but not extended
     * attr*0*=     sectioned and extended
     */
    private const string parameterAttributePattern = @"(?<attr>[a-zA-Z0-9\-]+)(?<section>\*\d+)?(?<ext>\*)?=";
    private static readonly Regex regexFirstParameterAttribute = new Regex(@"^" + parameterAttributePattern, RegexOptions.CultureInvariant | RegexOptions.Compiled);
    private static readonly Regex regexParameterAttribute = new Regex(@"(?<=;)\s*" + parameterAttributePattern, RegexOptions.CultureInvariant | RegexOptions.Compiled);
    private static readonly Regex regexNewline = new Regex(@"\s*([\r\n]+)\s*", RegexOptions.CultureInvariant | RegexOptions.Compiled | RegexOptions.Multiline);

    private static void ParseMatchedSectionAndExtension(Match match, out int? section, out bool extended)
    {
      if (string.IsNullOrEmpty(match.Groups["section"].Value))
        section = null;
      else
        section = int.Parse(match.Groups["section"].Value.Substring(1));

      extended = !string.IsNullOrEmpty(match.Groups["ext"].Value);
    }

    private class SectionedParameter {
      private static readonly char[] extendedInitialValueSplitter = new[] {'\''};

      private SortedList<int, string> sectionedValues = new SortedList<int, string>();
      private Dictionary<int, bool> isSectionExtended = new Dictionary<int, bool>();

      public void AddSection(bool extended, int? section, string value)
      {
        // section = null : extended but not sectioned
        sectionedValues[section ?? 0] = value;
        isSectionExtended[section ?? 0] = extended;
      }

      public string GetCombined(EncodingSelectionCallback selectFallbackCharset)
      {
        var charset = MimeMessage.DefaultCharset; // as default
        var ret = new StringBuilder();

        foreach (var pair in sectionedValues) {
          var section = pair.Key;
          var val = pair.Value;

          if (isSectionExtended[section]) {
            if (section == 0) {
              // extended-initial-value := [charset] "'" [language] "'" extended-other-values
              // extended-other-values := *(ext-octet / attribute-char)
              var vals = pair.Value.Split(extendedInitialValueSplitter);

              if (vals.Length == 3) {
                charset = vals[0].Length == 0
                  ? MimeMessage.DefaultCharset
                  : Charsets.FromString(vals[0], selectFallbackCharset);

                val = vals[2];
              }
            }

            val = PercentEncoding.GetDecodedString(val, charset, false);
          }

          ret.Append(val);
        }

        return ret.ToString();
      }
    }

    internal Dictionary<string, string> EnsureValueParsed(EncodingSelectionCallback selectFallbackCharset)
    {
      if (this.parameters != null)
        return this.parameters;

      var parms = new Dictionary<string, string>(StringComparer.OrdinalIgnoreCase);

      if (Value == null) {
        this.parameters = parms;
        return this.parameters;
      }

      var lineCollapsedValue = regexNewline.Replace(Value, " ").TrimEnd();

      if (lineCollapsedValue.EndsWith(";", StringComparison.Ordinal))
        lineCollapsedValue = lineCollapsedValue.Substring(0, lineCollapsedValue.Length - 1).TrimEnd();

      parms[string.Empty] = lineCollapsedValue; // if field has no parameters, set line-collapsed value

      var parmsSectioned = new Dictionary<string, SectionedParameter>(StringComparer.OrdinalIgnoreCase);

      Match lastMatch = null;
      string attr = null, val = null;
      int? section = null;
      bool extended = false;

      foreach (Match match in regexParameterAttribute.Matches(Value)) {
        if (lastMatch == null) {
          /* Header: field */
          var field = Value.Substring(0, match.Index - 1 /* ';' */).Trim();
          var m = regexFirstParameterAttribute.Match(field);

          if (m.Success) {
            /* Header: attr=value */
            parms[string.Empty] = string.Empty;

            attr = m.Groups["attr"].Value;
            val = field.Substring(m.Index + m.Length);

            ParseMatchedSectionAndExtension(m, out section, out extended);

            SetParameterValue(parms, parmsSectioned, attr, val, section, extended, selectFallbackCharset);
          }
          else {
            parms[string.Empty] = regexNewline.Replace(field, " "); // set line-collapsed value
          }
        }
        else {
          /* ;attr=value */
          var start = lastMatch.Index + lastMatch.Length;

          attr = lastMatch.Groups["attr"].Value;
          val = Value.Substring(start, (match.Index - 1 /* ';' */) - start);

          SetParameterValue(parms, parmsSectioned, attr, val, section, extended, selectFallbackCharset);
        }

        ParseMatchedSectionAndExtension(match, out section, out extended);

        lastMatch = match;
      }

      if (lastMatch != null) {
        attr = lastMatch.Groups["attr"].Value;
        val = Value.Substring(lastMatch.Index + lastMatch.Length);

        SetParameterValue(parms, parmsSectioned, attr, val, section, extended, selectFallbackCharset);
      }

      // combine sectioned parameters
      foreach (var pair in parmsSectioned) {
        parms[pair.Key] = pair.Value.GetCombined(selectFallbackCharset);
      }

      this.parameters = parms;

      return this.parameters;
    }

    private static void SetParameterValue(Dictionary<string, string> parms,
                                          Dictionary<string, SectionedParameter> parmsSectioned,
                                          string attr,
                                          string val,
                                          int? section,
                                          bool extended,
                                          EncodingSelectionCallback selectFallbackCharset)
    {
      val = val.Trim();

      if (val.StartsWith("\"", StringComparison.Ordinal) &&
          val.EndsWith("\"", StringComparison.Ordinal))
        // quoted
        val = MimeEncoding.Decode(val.Substring(1, val.Length - 2), selectFallbackCharset);

      if (extended || section.HasValue) {
        SectionedParameter sp;

        if (!parmsSectioned.TryGetValue(attr, out sp)) {
          sp = new SectionedParameter();

          parmsSectioned[attr] = sp;
        }

        sp.AddSection(extended, section, val);
      }
      else {
        parms[attr] = regexNewline.Replace(val, string.Empty);
      }
    }

    public override string ToString()
    {
      return string.Concat(name,
                           MimeHeaderBuilder.NameBodyFormattingDelimiter,
                           Value);
    }

    private readonly string name;
    private readonly MimeEncodingMethod encoding;
    private readonly Encoding charset;
    private Dictionary<string, string> parameters = null;
  }
}

==> Smdn.Formats.Mime-0.33.tar.bz2: Smdn.Formats.Mime-0.33/Smdn.Formats.Mime/Smdn.Formats.Mime/MimeHeader.cs <==
// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2012 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;

namespace Smdn.Formats.Mime {
  public class MimeHeader : IHeaderFieldBody, ICloneable {
    internal struct HeaderField {
      public ByteString Field {
        get { return field; }
      }

      public int IndexOfDelimiter {
        get { return indexOfDelimiter; }
      }

      public HeaderField(ByteString field, int indexOfDelimiter)
      {
        this.field = field;
        this.indexOfDelimiter = indexOfDelimiter;
      }

      private readonly ByteString field;
      private readonly int indexOfDelimiter;
    }

    public string Name {
      get { return name; }
    }

    public string Value {
      get; private set;
    }

    internal ByteString RawHeaderField {
      get { return headerField.Field; }
    }

    public MimeEncodingMethod Encoding {
      get { return encoding; }
    }

    public Encoding Charset {
      get { return charset; }
    }

    internal MimeHeader(HeaderField headerField, string decodedValue, MimeEncodingMethod encoding, Encoding charset)
    {
      this.headerField = headerField;
      this.name = headerField.Field.Substring(0, headerField.IndexOfDelimiter).TrimEnd().ToString();
      this.Value = decodedValue;
      this.encoding = encoding;
      this.charset = charset;
    }

    internal static MimeHeader CreateEmpty(string name)
    {
      var headerField = new HeaderField(ByteString.CreateImmutable(name + ":"),
                                        name.Length);

      return new MimeHeader(headerField, string.Empty, MimeEncodingMethod.None, Charsets.ISO8859_1);
    }

    public MimeHeader Clone()
    {
      var cloned = MemberwiseClone() as MimeHeader;

      if (this.parameters != null)
        cloned.parameters = new Dictionary<string, string>(this.parameters, this.parameters.Comparer);

      return cloned;
    }

    object ICloneable.Clone()
    {
      return Clone();
    }

    public bool IsNameEquals(string name)
    {
      return string.Equals(this.name, name, StringComparison.OrdinalIgnoreCase);
    }

    public string GetParameter(string attr)
    {
      if (attr == null)
        throw new ArgumentNullException("attr");
      if (attr.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString("attr");

      string val;

      if (EnsureValueParsed(null).TryGetValue(attr, out val))
        return val;
      else
        return null;
    }

    public KeyValuePair<string, string>[] GetAllParameters()
    {
      var ret = new List<KeyValuePair<string, string>>();

      foreach (var pair in EnsureValueParsed(null)) {
        if (pair.Key.Length == 0)
          continue; // if header field

        ret.Add(pair);
      }

      return ret.ToArray();
    }

    public void ForEachParameter(Action<KeyValuePair<string, string>> action)
    {
      if (action == null)
        throw new ArgumentNullException("action");

      foreach (var pair in EnsureValueParsed(null)) {
        if (pair.Key.Length == 0)
          continue; // if header field

        action(pair);
      }
    }

    public bool HasParameter(string attr)
    {
      if (attr == null)
        throw new ArgumentNullException("attr");
      if (attr.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString("attr");

      return EnsureValueParsed(null).ContainsKey(attr);
    }

    public string GetValueWithoutParameter()
    {
      string val;

      if (EnsureValueParsed(null).TryGetValue(string.Empty, out val))
        return val;
      else
        return null;
    }

    /*
     * attr=""      quoted
     * attr*=       extended
     * attr*0=      sectioned but not extended
     * attr*0*=     sectioned and extended
     */
    private const string parameterAttributePattern = @"(?<attr>[a-zA-Z0-9\-]+)(?<section>\*\d+)?(?<ext>\*)?=";
    private static readonly Regex regexFirstParameterAttribute = new Regex(@"^" + parameterAttributePattern, RegexOptions.CultureInvariant | RegexOptions.Compiled);
    private static readonly Regex regexParameterAttribute = new Regex(@"(?<=;)\s*" + parameterAttributePattern, RegexOptions.CultureInvariant | RegexOptions.Compiled);
    private static readonly Regex regexNewline = new Regex(@"\s*([\r\n]+)\s*", RegexOptions.CultureInvariant | RegexOptions.Compiled | RegexOptions.Multiline);

    private static void ParseMatchedSectionAndExtension(Match match, out int? section, out bool extended)
    {
      if (string.IsNullOrEmpty(match.Groups["section"].Value))
        section = null;
      else
        section = int.Parse(match.Groups["section"].Value.Substring(1));

      extended = !string.IsNullOrEmpty(match.Groups["ext"].Value);
    }

    private class SectionedParameter {
      private static readonly char[] extendedInitialValueSplitter = new[] {'\''};

      private SortedList<int, string> sectionedValues = new SortedList<int, string>();
      private Dictionary<int, bool> isSectionExtended = new Dictionary<int, bool>();

      public void AddSection(bool extended, int? section, string value)
      {
        // section = null : extended but not sectioned
        sectionedValues[section ?? 0] = value;
        isSectionExtended[section ?? 0] = extended;
      }

      public string GetCombined(EncodingSelectionCallback selectFallbackCharset)
      {
        var charset = MimeMessage.DefaultCharset; // as default
        var ret = new StringBuilder();

        foreach (var pair in sectionedValues) {
          var section = pair.Key;
          var val = pair.Value;

          if (isSectionExtended[section]) {
            if (section == 0) {
              // extended-initial-value := [charset] "'" [language] "'" extended-other-values
              // extended-other-values := *(ext-octet / attribute-char)
              var vals = pair.Value.Split(extendedInitialValueSplitter);

              if (vals.Length == 3) {
                charset = vals[0].Length == 0
                  ? MimeMessage.DefaultCharset
                  : Charsets.FromString(vals[0], selectFallbackCharset);

                val = vals[2];
              }
            }

            val = PercentEncoding.GetDecodedString(val, charset, false);
          }

          ret.Append(val);
        }

        return ret.ToString();
      }
    }

    internal Dictionary<string, string> EnsureValueParsed(EncodingSelectionCallback selectFallbackCharset)
    {
      if (this.parameters != null)
        return this.parameters;

      var parms = new Dictionary<string, string>(StringComparer.OrdinalIgnoreCase);

      if (Value == null) {
        this.parameters = parms;
        return this.parameters;
      }

      var lineCollapsedValue = regexNewline.Replace(Value, " ").TrimEnd();

      if (lineCollapsedValue.EndsWith(";", StringComparison.Ordinal))
        lineCollapsedValue = lineCollapsedValue.Substring(0, lineCollapsedValue.Length - 1).TrimEnd();

      parms[string.Empty] = lineCollapsedValue; // if field has no parameters, set line-collapsed value

      var parmsSectioned = new Dictionary<string, SectionedParameter>(StringComparer.OrdinalIgnoreCase);

      Match lastMatch = null;
      string attr = null, val = null;
      int? section = null;
      bool extended = false;

      foreach (Match match in regexParameterAttribute.Matches(Value)) {
        if (lastMatch == null) {
          /* Header: field */
          var field = Value.Substring(0, match.Index - 1 /* ';' */).Trim();
          var m = regexFirstParameterAttribute.Match(field);

          if (m.Success) {
            /* Header: attr=value */
            parms[string.Empty] = string.Empty;

            attr = m.Groups["attr"].Value;
            val = field.Substring(m.Index + m.Length);

            ParseMatchedSectionAndExtension(m, out section, out extended);

            SetParameterValue(parms, parmsSectioned, attr, val, section, extended, selectFallbackCharset);
          }
          else {
            parms[string.Empty] = regexNewline.Replace(field, " "); // set line-collapsed value
          }
        }
        else {
          /* ;attr=value */
          var start = lastMatch.Index + lastMatch.Length;

          attr = lastMatch.Groups["attr"].Value;
          val = Value.Substring(start, (match.Index - 1 /* ';' */) - start);

          SetParameterValue(parms, parmsSectioned, attr, val, section, extended, selectFallbackCharset);
        }

        ParseMatchedSectionAndExtension(match, out section, out extended);

        lastMatch = match;
      }

      if (lastMatch != null) {
        attr = lastMatch.Groups["attr"].Value;
        val = Value.Substring(lastMatch.Index + lastMatch.Length).TrimEnd();

        if (val.EndsWith(";", StringComparison.Ordinal))
          val = val.Substring(0, val.Length - 1);

        SetParameterValue(parms, parmsSectioned, attr, val, section, extended, selectFallbackCharset);
      }

      // combine sectioned parameters
      foreach (var pair in parmsSectioned) {
        parms[pair.Key] = pair.Value.GetCombined(selectFallbackCharset);
      }

      this.parameters = parms;

      return this.parameters;
    }

    private static void SetParameterValue(Dictionary<string, string> parms,
                                          Dictionary<string, SectionedParameter> parmsSectioned,
                                          string attr,
                                          string val,
                                          int? section,
                                          bool extended,
                                          EncodingSelectionCallback selectFallbackCharset)
    {
      val = val.Trim();

      if (val.StartsWith("\"", StringComparison.Ordinal) &&
          val.EndsWith("\"", StringComparison.Ordinal))
        // quoted
        val = MimeEncoding.Decode(val.Substring(1, val.Length - 2), selectFallbackCharset);

      if (extended || section.HasValue) {
        SectionedParameter sp;

        if (!parmsSectioned.TryGetValue(attr, out sp)) {
          sp = new SectionedParameter();

          parmsSectioned[attr] = sp;
        }

        sp.AddSection(extended, section, val);
      }
      else {
        parms[attr] = regexNewline.Replace(val, string.Empty);
      }
    }

    public override string ToString()
    {
      return headerField.Field.ToString();
    }

    private readonly HeaderField headerField;
    private readonly string name;
    private readonly MimeEncodingMethod encoding;
    private readonly Encoding charset;
    private Dictionary<string, string> parameters = null;
  }
}
